/* Ocamlyacc parser for CARL 
This code was original from MicroC and modified by:
Keir Lauritzen, Leon Song, Guy Yardeni, and Darren Hakimi

*/

%{
open Ast
let quote_remover a = String.sub a 1 ((String.length a) - 2);;

%}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT
%token PERIOD QUEST CARROT VERT LBRACK RBRACK
%token EQ NEQ LT LEQ GT GEQ TRUE FALSE AND OR
%token IF ELSEIF ELSE
%token RETURN BOOL VOID FLOAT
%token FOR WHILE CONCAT
%token RS FS NR NF RSTART RLENGTH
%token EOF
%token BEGIN END FUNCTION FSLASH STRING ARRAY_F ARRAY_S
%token RECORD RFIELD PATTERNS ACTIONS
%token RMINUS
%token <string> ID
%token <float> FLOAT_L
%token <string> STRING_L
%token <string> REGEX_STRING

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT NEG

%start program
%type <Ast.program> program

%%

program:
    fdecl_list rule_list EOF { { func_decls =List.rev $1; rules= List.rev $2 } }


fdecl_list:
    /* nothing */  { [] }
  | fdecl_list fdecl { $2 :: $1 }

fdecl:
   FUNCTION typ ID LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
     { { typ = $2;
   fname = $3;
   formals = $5;
   locals = List.rev $8;
   body = List.rev $9 } }

typ:
  | VOID { Void }
  | STRING { String_t }
  | FLOAT { Float }
  | ARRAY_F { Array_f }
  | ARRAY_S { Array_s }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    typ ID                   { [($1,$2)] }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
    typ ID assign_opt SEMI { ($1, $2, $3) }

assign_opt:
    /* nothing */  { Noexpr }
  | ASSIGN expr { $2 }



rule_list:
    /* nothing */  { [] }
  | rule_list rdecl { $2 :: $1 }

rdecl:
    pattern action { $1, $2 }

pattern:
    BEGIN { Begin }
    | END { End }
    | FSLASH regex_sequence FSLASH { RegexPattern($2) }

action:
    LBRACE vdecl_list stmt_list RBRACE  { List.rev $2, List.rev $3  }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    | expr SEMI { Expr $1 }
    | RETURN expr SEMI { Return($2) }
    | LBRACE stmt_list RBRACE { Block(List.rev $2) }
    | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
    | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7) }
    | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
        { For($3, $5, $7, $9) }
    | WHILE LPAREN expr RPAREN stmt { While($3, $5) }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
      STRING_L          { String_Lit(quote_remover($1)) }
    | FLOAT_L           { Float_Lit($1) }
    | LBRACK kvps_f RBRACK { Array_F_Lit($2) }
    | LBRACK kvps_s RBRACK { Array_S_Lit($2) }
    | ID               { Id($1) }
    | expr PLUS   expr { Binop($1, Add,   $3) }
    | expr MINUS  expr { Binop($1, Sub,   $3) }
    | expr TIMES  expr { Binop($1, Mult,  $3) }
    | expr DIVIDE expr { Binop($1, Div,   $3) }
    | expr EQ     expr { Binop($1, Equal, $3) }
    | expr NEQ    expr { Binop($1, Neq,   $3) }
    | expr LT     expr { Binop($1, Less,  $3) }
    | expr LEQ    expr { Binop($1, Leq,   $3) }
    | expr GT     expr { Binop($1, Greater, $3) }
    | expr GEQ    expr { Binop($1, Geq,   $3) }
    | expr AND    expr { Binop($1, And,   $3) }
    | expr OR     expr { Binop($1, Or,    $3) }
    | MINUS expr %prec NEG { Unop(Neg, $2) }
    | NOT expr         { Unop(Not, $2) }
    | ID ASSIGN expr   { Assign($1, $3) }
    | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
    | LPAREN expr RPAREN { $2 }
    | LBRACK RBRACK     { Call("create", [])}
    | ID LBRACK expr RBRACK { Retrieve($1, $3)}
    | ID LBRACK expr RBRACK ASSIGN expr {Array_Assign($1, $3, $6)}

kvps_f:
    kvp_f        { [$1] }
  | kvps_f COMMA kvp_f { $3 :: $1 }

kvps_s:
    kvp_s        { [$1] }
  | kvps_s COMMA kvp_s { $3 :: $1 }

kvp_f:
  STRING_L ASSIGN FLOAT_L { $1, $3 }

kvp_s:
  STRING_L ASSIGN STRING_L { $1, $3 }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }


/*Start of Regex*/

regex_sequence:
    regex {$1}
  | regex regex_sequence {$1^$2}

regex:
    REGEX_STRING {$1}
  | PERIOD {"."}
  | LPAREN regex_sequence RPAREN {"("^$2^")"}
  | LBRACK regex_set_sequence RBRACK {"["^$2^"]"}
  | regex QUEST {$1^"?"}
  | regex PLUS {$1^"+"}
  | regex TIMES {$1^"*"}
  | regex VERT  {$1^"|"}

regex_set:
    REGEX_STRING {$1}
  | REGEX_STRING RMINUS REGEX_STRING{$1^"-"^$3}
  | CARROT regex_set {"^"^$2}
  | LBRACK regex_set_sequence RBRACK {"["^$2^"]"}

regex_set_sequence:
    regex_set {$1}
  | regex_set regex_set_sequence {$1^$2}

/*End of Regex */
